;; Generated package description from ebdb.el
(define-package "ebdb" "0.6" "Contact management package" '((emacs "25.1") (cl-lib "0.5") (seq "2.15")) :url "https://github.com/girzel/ebdb" :keywords '("convenience" "mail"))
